from typing import Dict, Iterable, List, Optional, Sequence

import torch

from l5kit.cle.composite_metrics import SupportsCompositeMetricCompute
from l5kit.cle.metrics import SupportsMetricCompute
from l5kit.cle.validators import SupportsMetricValidate, ValidatorOutput
from l5kit.simulation.unroll import SimulationOutputCLE


class EvaluationPlan:
    """Evaluation plan describes a plan to evaluate metrics and run
    validators. It is composed by the list of metrics that should be computed
    as well as the list of validators that will run. It checks for consistency
    of the plan (validators depend on metrics).

    .. note:: Please note that the intervention_validators argument, that
              specifies a list of interventions will stop the validation
              of all other metrics if any validator specified in this list
              is triggered.

    :param metrics: list of the metrics to compute
    :param validators: list of validators to compute
    :param composite_metrics: list of composite metrics to compute
    :param intervention_validators: list of validators that are considered
                                    interventions.
    """

    def __init__(self, metrics: Iterable[SupportsMetricCompute],
                 validators: Optional[Iterable[SupportsMetricValidate]] = None,
                 composite_metrics: Optional[Iterable[SupportsCompositeMetricCompute]] = None,
                 intervention_validators: Optional[List[str]] = None):
        self.metrics = metrics
        self.validators = validators or []
        self.composite_metrics = composite_metrics or []
        self.intervention_validators = intervention_validators or []
        self._validate_plan()

    def metrics_dict(self) -> Dict[str, SupportsMetricCompute]:
        """Get the metric names and metrics from the plan."""
        return {m.metric_name: m for m in self.metrics}

    def validators_dict(self) -> Dict[str, SupportsMetricValidate]:
        """Get the validator names and validators from the plan."""
        return {v.validator_name: v for v in self.validators}

    def composite_metrics_dict(self) -> Dict[str, SupportsCompositeMetricCompute]:
        """Get the composite metric names and composite metrics from the plan."""
        return {cm.composite_metric_name: cm for cm in self.composite_metrics}

    def _validate_plan(self) -> None:
        """Check for consistency of the plan, all metrics required by
        validators should be specified in the plan.
        """
        # Check for repetition of metrics and validators
        metric_names = [m.metric_name for m in self.metrics]
        validator_names = [v.validator_name for v in self.validators]
        composite_metric_names = [cm.composite_metric_name for cm in self.composite_metrics]

        if len(set(metric_names)) != len(metric_names):
            raise RuntimeError("You cannot have repeated metric names.")

        if len(set(validator_names)) != len(validator_names):
            raise RuntimeError("You cannot have repeated validator names.")

        if len(set(composite_metric_names)) != len(composite_metric_names):
            raise RuntimeError("You cannot have repeated composite metric names.")

        # Check if we have all validators specified in the intervention list
        for vname in self.intervention_validators:
            if vname not in validator_names:
                raise RuntimeError(f"Validator '{vname}' not found in validators list.")

        # Check for consistency of the validators, if we have all required
        # metrics especified.
        metric_calculators = self.metrics_dict()
        for validator in self.validators:
            for metric_requirement in validator.requires_metric:
                if metric_requirement not in metric_calculators:
                    raise RuntimeError(f"Metric '{metric_requirement}' required "
                                       f"by validator '{validator.validator_name}'.")

        # Check for consistency of the composite metrics, if we have all required
        # metrics and validators specified.
        validators_specified = self.validators_dict()
        for cm in self.composite_metrics:
            # Check for metric requirements
            for metric_requirement in cm.requires_metric:
                if metric_requirement not in metric_calculators:
                    raise RuntimeError(f"Metric '{metric_requirement}' required "
                                       f"by composite metric '{cm.composite_metric_name}'.")
            # Check for validator requirements
            for validator_requirement in cm.requires_validator:
                if validator_requirement not in validators_specified:
                    raise RuntimeError(f"Validator '{validator_requirement}' required "
                                       f"by composite metric '{cm.composite_metric_name}'.")

    def evaluate(self, simulation_output: SimulationOutputCLE) -> Dict[str, torch.Tensor]:
        """Execute the evaluation (metric computation) on the scene.

        :param simulation_output: output from the closed-loop simulator
        :returns: results from all metrics on a dictionary
        """
        results: Dict[str, torch.Tensor] = {}
        for metric_calculator in self.metrics:
            metric_result = metric_calculator.compute(simulation_output)
            results[metric_calculator.metric_name] = metric_result
        return results

    def evaluate_composite(self, simulation_output: SimulationOutputCLE,
                           scene_metrics: Dict[str, torch.Tensor],
                           scene_validation: Dict[str, ValidatorOutput]) -> Dict[str, float]:
        """Execute the evaluation of the composite metrics on the scene.

        :param simulation_output: output from the closed-loop simulator
        :param scene_metrics: metric results indexed by the metric name
        :param scene_validation: outputs from validator indexed by the validation name
        :return: results from the composite metrics indexed by the composite metric name
        """
        results: Dict[str, float] = {}
        for cm in self.composite_metrics:
            # Composite metrics should only see metrics and validators they require
            required_metrics = cm.requires_metric
            required_validators = cm.requires_validator
            # Filter to keep required metrics only
            scene_metrics_required = {
                metric_name: scene_metrics[metric_name]
                for metric_name in required_metrics
            }
            # Filter to keep required validators only
            scene_validators_required = {
                validator_name: scene_validation[validator_name]
                for validator_name in required_validators
            }

            cm_result = cm.compute(scene_metrics_required,
                                   scene_validators_required,
                                   simulation_output)
            results[cm.composite_metric_name] = cm_result
        return results

    def process_interventions(self,
                              results: Dict[str, ValidatorOutput]) -> Dict[str, ValidatorOutput]:
        """This method will process the the validator results accordingly
        to the validators defined as interventions. If any validator, that
        is also an intervention is triggered, it will reset all other validators.

        :param results: resuls from the validation
        :returns: updated the results
        """
        min_intervention_name = None
        min_intenvention_frame = float('inf')

        # For each intervention validator in the list
        for ivname in self.intervention_validators:
            voutput: ValidatorOutput = results[ivname]
            failed_frames: List[int] = voutput.failed_frames
            if len(failed_frames) <= 0:
                continue

            inner_min_failed_frame = min(failed_frames)
            if inner_min_failed_frame < min_intenvention_frame:
                min_intenvention_frame = inner_min_failed_frame
                min_intervention_name = ivname

        # If an intervention happened, all other validators are reset
        # as if they were not triggered.
        if min_intervention_name is not None:
            new_results = {ivname: ValidatorOutput(True, [])
                           for ivname in results.keys()}
            new_results[min_intervention_name] = results[min_intervention_name]
            return new_results
        else:
            return results

    def validate(self, scene_metrics: Dict[str, torch.Tensor],
                 simulation_output: SimulationOutputCLE) -> Dict[str, ValidatorOutput]:
        """Execute the validation (validators) on all metric results.

        :param scene_metrics: the result for the metrics computation
        :param simulation_output: output from the closed-loop simulator
        :returns: the result of all validators
        """
        results: Dict[str, ValidatorOutput] = {}
        for metric_validator in self.validators:
            # Validators should only see metrics they require
            required_metrics = metric_validator.requires_metric
            scene_metrics_required = {
                metric_name: scene_metrics[metric_name] for metric_name in required_metrics
            }
            validator_result = metric_validator.validate(scene_metrics_required,
                                                         simulation_output)
            results[metric_validator.validator_name] = validator_result

        # If we should stop at interventions, given a list of interventions
        if len(self.intervention_validators) > 0:
            results = self.process_interventions(results)

        return results


class ClosedLoopEvaluator:
    """The closed loop evaluator executes a evaluation plan and keep
    track of histograms, failed scenes, etc.

    :param evaluation_plan: the specified evaluation plan
    """

    #: Results of the metrics indexed by the scene id
    scene_metric_results: Dict[int, Dict[str, torch.Tensor]]
    #: Results from the validation results indexed by the scene id
    scene_validation_results: Dict[int, Dict[str, ValidatorOutput]]
    #: Results from the composite metrics indexed by the scene id
    scene_composite_metric_results: Dict[int, Dict[str, float]]

    def __init__(self, evaluation_plan: EvaluationPlan):
        self.evaluation_plan = evaluation_plan
        self.scene_validation_results = {}
        self.scene_metric_results = {}
        self.scene_composite_metric_results = {}

    def reset(self) -> None:
        """Resets the computed stats."""
        self.scene_validation_results = {}
        self.scene_metric_results = {}
        self.scene_composite_metric_results = {}

    def metric_results(self) -> Dict[int, Dict[str, torch.Tensor]]:
        """Return the computed metric results.

        :return: a dictionary indexed by scene with metric name
                 and results.
        """
        return self.scene_metric_results

    def validation_results(self) -> Dict[int, Dict[str, ValidatorOutput]]:
        """Return the computed validator results.

        :return: a dictionary indexed by scene with validator name
                 and results.
        """
        return self.scene_validation_results

    def composite_metric_results(self) -> Dict[int, Dict[str, float]]:
        """Return the computed composite metric results.

        :return: a dictionary indexed by scene with composite metric name
                 and results.
        """
        return self.scene_composite_metric_results

    def evaluate(self, simulation_outputs: Sequence[SimulationOutputCLE]) -> None:
        """Executes the evaluation plan on all outputs from the simulator.

        :param simulation_outputs: the outputs from the simulator
        """
        # TODO(perone): local parallelization can be done here
        for simulation_output in simulation_outputs:
            scene_id = simulation_output.get_scene_id()

            # Run metric calculation, metrics_for_scene here is a dict
            # where the tensor is the same size of the scene
            metrics_for_scene: Dict[str, torch.Tensor] = \
                self.evaluation_plan.evaluate(simulation_output)
            self.scene_metric_results[scene_id] = metrics_for_scene

            # Run validators
            validation_for_scene = self.evaluation_plan.validate(metrics_for_scene,
                                                                 simulation_output)
            self.scene_validation_results[scene_id] = validation_for_scene

            # Run composite metrics: these are metrics that depend on the result
            # of validators and metrics
            cm_for_scene = self.evaluation_plan.evaluate_composite(simulation_output,
                                                                   metrics_for_scene,
                                                                   validation_for_scene)
            self.scene_composite_metric_results[scene_id] = cm_for_scene



from .random_generator import (GaussianRandomGenerator, LambdaRandomGenerator, RandomGenerator, ReplayRandomGenerator,
                               UniformRandomGenerator)


__all__ = ["GaussianRandomGenerator", "LambdaRandomGenerator",
           "RandomGenerator", "ReplayRandomGenerator", "UniformRandomGenerator"]



from abc import ABC, abstractmethod
from typing import Callable, Optional

import numpy as np


class RandomGenerator(ABC):
    @abstractmethod
    def _sample(self) -> np.ndarray:
        """The sampling function that used to generate random samples.
        :return: samples generated by defined sampling function
        :rtype: np.ndarray
        """
        pass

    def __call__(self) -> np.ndarray:
        """A function that calls sampling function and returns the generated sample.
        :return: samples generated by defined sampling function
        :rtype: np.ndarray
        """
        return self._sample()


class LambdaRandomGenerator(RandomGenerator):
    """
    LambdaRandomGenerator generates values by calling the supplied function.
    Note that instead of this you could also just use this supplied function itself.
    """

    def __init__(self, sampling_function: Callable[[], np.ndarray]):
        self.sampling_function = sampling_function

    def _sample(self) -> np.ndarray:
        return self.sampling_function()


class ReplayRandomGenerator(RandomGenerator):
    """
    ReplayRandomGenerator generates values by sequentially returning the values in the supplied
    array. It does not loop or reset automatically.
    """

    def __init__(self, values: np.ndarray):
        self._values = values
        self._idx = 0

    def _sample(self) -> np.ndarray:
        if self._idx == len(self._values):
            raise IndexError("ReplayRandomGenerator is out of values for index={}".format(self._idx))
        val = np.array(self._values[self._idx])
        self._idx += 1
        return val


class GaussianRandomGenerator(RandomGenerator):
    """
    GaussianRandomGenerator draws samples from a normal distribution with specified mean and standard deviation.
    Note that this gaussian can be multidimensional.
    """

    def __init__(self, mean: np.ndarray, std: np.ndarray, random_seed: Optional[int] = 0):
        self.mean = np.array(mean)
        self.std = np.array(std)
        self.rng = np.random.default_rng(random_seed)

    def _sample(self) -> np.ndarray:
        return self.rng.normal(self.mean, self.std)


class UniformRandomGenerator(RandomGenerator):
    """
    UniformRandomGenerator draws samples from a uniform distribution over the half-open interval [low, high).
    """

    def __init__(self, low: np.ndarray, high: np.ndarray):
        self.low = low
        self.high = high

    def _sample(self) -> np.ndarray:
        return np.random.uniform(self.low, self.high)



import math
from typing import Tuple

import numpy as np
from scipy import optimize

from l5kit.geometry import angular_distance


def fit_ackerman_model_approximate(
    gx: np.ndarray,
    gy: np.ndarray,
    gr: np.ndarray,
    gv: np.ndarray,
    wx: np.ndarray,
    wy: np.ndarray,
    wr: np.ndarray,
    wv: np.ndarray,
    wgx: np.ndarray,
    wgy: np.ndarray,
    wgr: np.ndarray,
    wgv: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Fits feasible ackerman-steering trajectory to groundtruth control points.
    Groundtruth is represented as 4 input numpy arrays ``(gx, gy, gr, gv)``
    each of size ``(N,)`` representing position, rotation and velocity at time ``i``.
    Returns 4 arrays ``(x, y, r, v)`` each of shape ``(N,)`` - the optimal trajectory.
    The solution is found as minimization of the following non-linear least squares problem:
    minimize F(x, y, r, v) = F_ground_cost(x, y, r, v) + F_kinematics_cost(x, y, r, v) where
    F_ground_cost(x, y, r, v) = 0.5 * sum(
    (wgx[i] * (x[i] - gx[i])) ** 2 +
    (wgy[i] * (y[i] - gy[i])) ** 2 +
    (wgr[i] * (r[i] - gr[i])) ** 2 +
    (wgv[i] * (v[i] - gv[i])) ** 2,
    i = 0 ... N-1)
    and
    F_kinematics_cost(x, y, r, v) = 0.5 * sum(
    (wx * (x[i] + cos(r[i]) * v[i] - x[i+1])) ** 2 +
    (wy * (y[i] + sin(r[i]) * v[i] - y[i+1])) ** 2 +
    (wr * (r[i] - r[i+1])) ** 2 +
    (wv * (v[i] - v[i+1])) ** 2,
    i = 0 ... N-2)
    Weights wg* control adherance to the control points while
    weights w* control obeying of underlying kinematic motion constrains.

    :return: 4 arrays (x, y, r, v) each of shape (N,), the optimal trajectory.
    """

    N = len(gx)

    w = np.hstack([wgx, wgy, wgr, wgv, wx, wy, wr, wv])

    def residuals(xyrv: np.ndarray) -> np.ndarray:
        x, y, r, v = np.split(xyrv, 4)

        x1, x2 = x[0:N - 1], x[1:N]
        y1, y2 = y[0:N - 1], y[1:N]
        r1, r2 = r[0:N - 1], r[1:N]
        v1, v2 = v[0:N - 1], v[1:N]

        return w * np.hstack(
            [
                x - gx,
                y - gy,
                angular_distance(r, gr),
                v - gv,
                np.append(x1 + np.cos(r1) * v1 - x2, 0),
                np.append(y1 + np.sin(r1) * v1 - y2, 0),
                np.append(angular_distance(r1, r2), 0),
                np.append(v1 - v2, 0),
            ]
        )

    # jacobian of residuals
    def jacobian(xyrv: np.ndarray) -> np.ndarray:
        x, y, r, v = np.split(xyrv, 4)

        z = np.zeros((N, N))
        e = np.eye(N, N)
        e0 = np.block([[np.eye(N - 1, N - 1), np.zeros((N - 1, 1))], [np.zeros((1, N))]])
        e1 = np.block([[np.zeros((N - 1, 1)), np.eye(N - 1, N - 1)], [np.zeros((1, N))]])

        A = np.block(
            [
                [e, z, z, z],
                [z, e, z, z],
                [z, z, e, z],
                [z, z, z, e],
                [e0 - e1, z, -np.sin(r) * v * e0, np.cos(r) * e0],
                [z, e0 - e1, np.cos(r) * v * e0, np.sin(r) * e0],
                [z, z, e0 - e1, z],
                [z, z, z, e0 - e1],
            ]
        )

        return w[:, None] * A

    # Gauss–Newton algorithm
    xyrv = np.hstack([gx, gy, gr, gv])
    for _ in range(5):
        xyrv = xyrv - np.linalg.lstsq(jacobian(xyrv), residuals(xyrv), rcond=None)[0]
    x, y, r, v = np.split(xyrv, 4)
    return x, y, r, v


def fit_ackerman_model_exact(
    x0: np.ndarray,
    y0: np.ndarray,
    r0: np.ndarray,
    v0: np.ndarray,
    gx: np.ndarray,
    gy: np.ndarray,
    gr: np.ndarray,
    gv: np.ndarray,
    wgx: np.ndarray,
    wgy: np.ndarray,
    wgr: np.ndarray,
    wgv: np.ndarray,
    ws: float = 5.0,
    wa: float = 5.0,
    min_acc: float = -0.3,  # min acceleration: -3 mps2
    max_acc: float = 0.3,   # max acceleration: 3 mps2
    min_steer: float = -math.radians(45) * 0.1,  # max yaw rate: 45 degrees per second
    max_steer: float = math.radians(45) * 0.1,   # max yaw rate: 45 degrees per second
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Fits feasible ackerman-steering trajectory to groundtruth control points.
    Groundtruth is represented as 4 numpy arrays ``(gx, gy, gr, gv)``
    each of shape ``(N,)`` representing position, rotation and velocity at time i.
    Returns 4 arrays ``(x, y, r, v)`` each of shape ``(N,)`` - the optimal trajectory.

    The solution is found as minimisation of the following non-linear least squares problem:
    ::
    minimize F(steer, acc) = 0.5 * sum(
    (wgx[i] * (x[i] - gx[i])) ** 2 +
    (wgy[i] * (y[i] - gy[i])) ** 2 +
    (wgr[i] * (r[i] - gr[i])) ** 2 +
    (wgv[i] * (v[i] - gv[i])) ** 2 +
    (ws * steer[i]) ** 2 +
    (wa * acc[i]) ** 2)
    i = 1 ... N)
    subject to following unicycle motion model equations:
    x[i+1] = x[i] + cos(r[i]) * v[i]
    y[i+1] = y[i] + sin(r[i]) * v[i]
    r[i+1] = r[i] + steer[i]
    v[i+1] = v[i] + acc[i]
    min_steer < steer[i] < max_steer
    min_acc < acc[i] < max_acc
    for i = 0 .. N
    Weights ``wg*`` control adherence to the control points
    In a typical usecase ``wgx = wgy = 1`` and ``wgr = wgv = 0``

    :return: 4 arrays ``(x, y, r, v)`` each of shape ``(N,)``- the optimal trajectory.
    """
    N = len(gx)

    wsteer_acc = np.hstack([ws * np.ones(N), wa * np.ones(N)])

    def control2position(steer_acc: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        steer, acc = np.split(steer_acc, 2)
        r = r0 + np.cumsum(steer)
        v = v0 + np.cumsum(acc)
        x = x0 + np.cumsum(np.cos(r) * v)
        y = y0 + np.cumsum(np.sin(r) * v)
        return x, y, r, v

    def residuals(steer_acc: np.ndarray) -> np.ndarray:
        x, y, r, v = control2position(steer_acc)
        return np.hstack(
            [
                wgx * (x - gx),
                wgy * (y - gy),
                wgr * angular_distance(r, gr),
                wgv * (v - gv),
                wsteer_acc * steer_acc,
            ]
        )

    def jacobian(steer_acc: np.ndarray) -> np.ndarray:
        x, y, r, v = control2position(steer_acc)

        Jr1, Jr2, Jr3, Jr4 = np.zeros(N), np.zeros(N), np.zeros(N), np.zeros(N)
        Jv1, Jv2, Jv3, Jv4 = np.zeros(N), np.zeros(N), np.zeros(N), np.zeros(N)
        J = np.zeros((2 * N, 4 * N))

        for i in range(N - 1, -1, -1):
            Jr1[i:] = Jr1[i:] - np.sin(r[i]) * v[i] * wgx[i:]
            Jr2[i:] = Jr2[i:] + np.cos(r[i]) * v[i] * wgy[i:]
            Jv1[i:] = Jv1[i:] + np.cos(r[i]) * wgx[i:]
            Jv2[i:] = Jv2[i:] + np.sin(r[i]) * wgy[i:]
            Jr3[i:] = wgr[i:]
            Jv4[i:] = wgv[i:]

            J[i, :] = np.hstack([Jr1, Jr2, Jr3, Jr4])
            J[N + i, :] = np.hstack([Jv1, Jv2, Jv3, Jv4])

        return np.vstack([J.T, np.eye(N + N) * wsteer_acc])

    min_bound = np.concatenate((min_steer * np.ones(N), min_acc * np.ones(N)))
    max_bound = np.concatenate((max_steer * np.ones(N), max_acc * np.ones(N)))
    result = optimize.least_squares(residuals, np.zeros(2 * N), jacobian, (min_bound, max_bound))

    x, y, r, v = control2position(result["x"])
    steer, acc = result["x"][:N], result["x"][N:]
    return x, y, r, v, acc, steer



from typing import List, Optional

import numpy as np

from .box_rasterizer import BoxRasterizer
from .rasterizer import Rasterizer
from .render_context import RenderContext
from .semantic_rasterizer import SemanticRasterizer


class SemBoxRasterizer(Rasterizer):
    """Combine a Semantic Map and a Box Rasterizers into a single class
    """

    def __init__(
            self,
            render_context: RenderContext,
            filter_agents_threshold: float,
            history_num_frames: int,
            semantic_map_path: str,
            world_to_ecef: np.ndarray,
            render_ego_history: bool = True,
    ):
        super(SemBoxRasterizer, self).__init__()
        self.render_context = render_context
        self.raster_size = render_context.raster_size_px
        self.pixel_size = render_context.pixel_size_m
        self.ego_center = render_context.center_in_raster_ratio
        self.filter_agents_threshold = filter_agents_threshold
        self.history_num_frames = history_num_frames

        self.box_rast = BoxRasterizer(render_context, filter_agents_threshold, history_num_frames, render_ego_history)
        self.sem_rast = SemanticRasterizer(render_context, semantic_map_path, world_to_ecef)

    def rasterize(
            self,
            history_frames: np.ndarray,
            history_agents: List[np.ndarray],
            history_tl_faces: List[np.ndarray],
            agent: Optional[np.ndarray] = None,
    ) -> np.ndarray:
        im_out_box = self.box_rast.rasterize(history_frames, history_agents, history_tl_faces, agent)
        im_out_sem = self.sem_rast.rasterize(history_frames, history_agents, history_tl_faces, agent)
        return np.concatenate([im_out_box, im_out_sem], -1)

    def to_rgb(self, in_im: np.ndarray, **kwargs: dict) -> np.ndarray:
        im_out_box = self.box_rast.to_rgb(in_im[..., :-3], **kwargs)
        im_out_sem = self.sem_rast.to_rgb(in_im[..., -3:], **kwargs)
        # merge the two together
        mask_box = np.any(im_out_box > 0, -1)
        im_out_sem[mask_box] = im_out_box[mask_box]
        return im_out_sem

    def num_channels(self) -> int:
        return self.box_rast.num_channels() + self.sem_rast.num_channels()



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional, Union


L5KIT_DATA_FOLDER_ENV_KEY = "L5KIT_DATA_FOLDER"


class DataManager(ABC):
    @abstractmethod
    def require(self, key: str) -> str:
        pass


class LocalDataManager(DataManager):
    """LocalDataManager allows you to require data to be present in the subpath of a specific folder.

    Example:
    Your data folder is set to ``"/tmp/my-data-folder"``, and you
    call ``local_data_manager.require("path/to/image.jpg")``, it would check if
    ``"/tmp/my-data-folder/path/to/image.jpg"`` exists, and if so return that complete path
    (``"/tmp/my-data-folder/path/to/image.jpg"``), otherwise it raises an error.

    In order of precedence, the local data folder is set by
      1. Passing in the path to the constructor of ``LocalDataManager``
      2. Setting the ``L5KIT_DATA_FOLDER`` environment variable.
    """

    def __init__(self, local_data_folder: Optional[Union[str, Path]] = None):
        if local_data_folder is None:
            if L5KIT_DATA_FOLDER_ENV_KEY in os.environ:
                local_data_folder = os.environ[L5KIT_DATA_FOLDER_ENV_KEY]
            else:
                raise ValueError(
                    f"{L5KIT_DATA_FOLDER_ENV_KEY} has not been set and you passed None to this call."
                    "either set the env variable or pass a valid path to this call"
                )

        self.root_folder = Path(local_data_folder)

    def require(self, key: str) -> str:
        """Require checks whether the file with the given key is present in the local data folder, if it is not it
        raises an error. Returns the path to the file otherwise.

        Arguments:
            key (str): Path from the data folder where the file or folder should be present.

        Returns:
            str -- Filepath including the data folder where required key is present.
        """
        local_path = self.root_folder / key
        local_path_str = str(local_path)

        if local_path.exists():
            return local_path_str
        else:
            raise FileNotFoundError(f"{key} is not present in local data folder {self.root_folder}")



import argparse
import multiprocessing
import os
import pprint
import sys
import warnings
from collections import Counter, defaultdict
from functools import partial
from multiprocessing import cpu_count, Pool
from pathlib import Path
from tempfile import gettempdir
from typing import Tuple
from uuid import uuid4

import numpy as np
import zarr
from prettytable import PrettyTable
from tqdm import tqdm

from l5kit.data import ChunkedDataset
from l5kit.data.filter import _get_label_filter  # TODO expose this without digging
from l5kit.geometry import angular_distance


if sys.platform == "darwin":
    multiprocessing.set_start_method("fork", force=True)  # this fixes loop in python 3.8 on MacOS

if sys.platform not in ["win32", "cygwin"]:
    os.environ["BLOSC_NOLOCK"] = "1"  # this is required for multiprocessing
else:
    warnings.warn(
        "Windows detected. BLOSC_NOLOCK has not been set as it causes memory leaks on Windows."
        "However, writing the mask with this config may be inconsistent."
    )

TH_YAW_DEGREE = 30
TH_EXTENT_RATIO = 1.1
TH_DISTANCE_AV = 50


def in_consecutive_frame(frame_idx: int, past_frame_idx: int) -> bool:
    return bool(frame_idx == past_frame_idx + 1)


def in_av_distance(av_translation: np.ndarray, agent_centroid: np.ndarray, th: float) -> bool:
    return bool(np.linalg.norm(av_translation[:2] - agent_centroid) < th)


def in_angular_distance(yaw1: np.ndarray, yaw2: np.ndarray, th: float) -> bool:
    """
    Check if the absolute distance in degrees is under the given threshold
    """

    abs_angular_distance_degrees = abs(angular_distance(yaw2, yaw1)) * 180 / np.pi
    return bool(abs_angular_distance_degrees < th)


def in_extent_ratio(extent1: np.ndarray, extent2: np.ndarray, th: float) -> bool:
    """
    Compute the two areas and then the ratio. The ratio is in the range [1, inf)
    """
    area_1 = extent1[0] * extent1[1]
    area_2 = extent2[0] * extent2[1]
    if area_1 < 0.01 or area_2 < 0.01:  # these are clearly errors (less than 1cm2)
        return False
    ratio = area_1 / area_2 if area_1 > area_2 else area_2 / area_1
    return bool(ratio < th)


def update_mask(mask: np.ndarray, agent_list: list) -> None:
    for idx_el, (frame_idx, mask_idx, agent) in enumerate(agent_list):
        mask[mask_idx][0] = idx_el  # past information
        mask[mask_idx][1] = len(agent_list) - idx_el - 1  # future information


def get_valid_agents(
        frames_range: np.ndarray,
        dataset: ChunkedDataset,
        th_agent_filter_probability_threshold: float,
        th_yaw_degree: float,
        th_extent_ratio: float,
        th_distance_av: float,
) -> Tuple[np.ndarray, Counter, tuple]:
    """
    Two types of filters are implemented:
    POINT-WISE: only the current state is considered
    COUPLE-WISE: 2 states considered (new and last added)

    Return a boolean np.array with the same shape of agents and a counter of report
    """
    frames = dataset.frames[slice(*frames_range)]
    agents_range_start = frames[0]["agent_index_interval"][0]
    agents_range_end = frames[-1]["agent_index_interval"][1]

    agents = dataset.agents[agents_range_start:agents_range_end]
    frames["agent_index_interval"] -= agents_range_start  # sync frame and agents again

    agents_dict = defaultdict(list)

    # for every agent in the .zarr -> (available_past_frame, available_future_frames) using this agent_threshold
    # this means a single mask can be used to generate all configurations of future and past frames
    agents_mask = np.zeros((len(agents), 2), dtype=np.uint32)

    report: Counter = Counter()

    # filter here for point-wise to speed up
    of_interest = _get_label_filter(agents["label_probabilities"], th_agent_filter_probability_threshold)
    global_agent_idx = -1
    for frame_idx in range(len(frames)):
        frame = frames[frame_idx]
        agents_frame = agents[slice(*(frame["agent_index_interval"]))]

        for agent in agents_frame:
            global_agent_idx += 1
            # store info we need for couple and sequence filters
            agents_dict[agent["track_id"]].append((frame_idx, global_agent_idx, agent))

            # ==== POINT-WISE FILTERS
            if not of_interest[global_agent_idx]:
                update_mask(agents_mask, agents_dict[agent["track_id"]][:-1])
                report["reject_th_agent_filter_probability_threshold"] += 1
                agents_dict[agent["track_id"]] = []
                continue

            if not in_av_distance(frame["ego_translation"], agent["centroid"], th_distance_av):
                update_mask(agents_mask, agents_dict[agent["track_id"]][:-1])
                report["reject_th_AV_distance"] += 1
                agents_dict[agent["track_id"]] = []
                continue

            # ==== COUPLE-WISE FILTERS
            if len(agents_dict[agent["track_id"]]) > 1:
                p_frame_idx, p_global_agent_idx, p_agent = agents_dict[agent["track_id"]][-2]  # get prev element

                if not in_consecutive_frame(frame_idx, p_frame_idx):
                    update_mask(agents_mask, agents_dict[agent["track_id"]][:-1])
                    report["reject_th_hole"] += 1
                    agents_dict[agent["track_id"]] = agents_dict[agent["track_id"]][-1:]
                    continue
                if not in_angular_distance(p_agent["yaw"], agent["yaw"], th_yaw_degree):
                    update_mask(agents_mask, agents_dict[agent["track_id"]][:-1])
                    report["reject_th_yaw"] += 1
                    agents_dict[agent["track_id"]] = agents_dict[agent["track_id"]][-1:]
                    continue
                if not in_extent_ratio(p_agent["extent"], agent["extent"], th_extent_ratio):
                    update_mask(agents_mask, agents_dict[agent["track_id"]][:-1])
                    report["reject_th_extent"] += 1
                    agents_dict[agent["track_id"]] = agents_dict[agent["track_id"]][-1:]
                    continue

    # update what is left inside the dict
    for track_id, agent_list in agents_dict.items():
        update_mask(agents_mask, agent_list)
        agents_dict[track_id] = []

    report["total_reject"] = sum([v for v in report.values()])
    report["total_agent_frames"] = len(agents_mask)
    return agents_mask, report, (agents_range_start, agents_range_end)


def select_agents(
        zarr_dataset: ChunkedDataset,
        th_agent_prob: float,
        th_yaw_degree: float,
        th_extent_ratio: float,
        th_distance_av: float,
) -> None:
    """
    Filter agents from zarr INPUT_FOLDER according to multiple thresholds and store a boolean array of the same shape.
    """
    agents_mask_path = Path(zarr_dataset.path) / f"agents_mask/{th_agent_prob}"

    if agents_mask_path.exists():
        raise FileExistsError(f"{th_agent_prob} exists already! only one is supported!")

    frame_index_intervals = zarr_dataset.scenes["frame_index_interval"]

    # build a partial with all args except the first one (will be passed by threads)
    get_valid_agents_partial = partial(
        get_valid_agents,
        dataset=zarr_dataset,
        th_agent_filter_probability_threshold=th_agent_prob,
        th_yaw_degree=th_yaw_degree,
        th_extent_ratio=th_extent_ratio,
        th_distance_av=th_distance_av,
    )

    try:
        root = zarr.open(zarr_dataset.path, mode="a")
        root.create_group("agents_mask")
    except ValueError:
        pass  # group is already there

    agents_mask = zarr.open_array(
        str(agents_mask_path),
        mode="w",
        shape=(len(zarr_dataset.agents), 2),
        chunks=(10000,),
        dtype=np.uint32,
        synchronizer=zarr.ProcessSynchronizer(Path(gettempdir()) / f"ag_mask_{str(uuid4())}.sync"),
    )

    report: Counter = Counter()
    print("starting pool...")
    with Pool(cpu_count()) as pool:
        tasks = tqdm(enumerate(pool.imap_unordered(get_valid_agents_partial, frame_index_intervals)))
        for idx, (mask, count, agents_range) in tasks:
            report += count
            agents_mask[agents_range[0]: agents_range[1]] = mask
            tasks.set_description(f"{idx + 1}/{len(frame_index_intervals)}")
        print("collecting results..")

    agents_cfg = {
        "th_agent_filter_probability_threshold": th_agent_prob,
        "th_yaw_degree": th_yaw_degree,
        "th_extent_ratio": th_extent_ratio,
        "th_distance_av": th_distance_av,
    }
    # print report
    pp = pprint.PrettyPrinter(indent=4)
    print(f"start report for {zarr_dataset.path}")
    pp.pprint({**agents_cfg, **report})

    future_steps = [0, 10, 30, 50]
    past_steps = [0, 10, 30, 50]
    agents_mask_np = np.asarray(agents_mask)

    table = PrettyTable(field_names=["past/future"] + [str(step) for step in future_steps])
    for step_p in tqdm(past_steps, desc="computing past/future table"):
        row = [step_p]
        for step_f in future_steps:
            past_mask = agents_mask_np[:, 0] >= step_p
            future_mask = agents_mask_np[:, 1] >= step_f
            row.append(np.sum(past_mask * future_mask))
        table.add_row(row)
    print(table)
    print(f"end report for {zarr_dataset.path}")
    print("==============================")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--input_folders", nargs="+", type=str, required=True, help="zarr path")
    parser.add_argument("--th_agent_prob", type=float, required=True, help="perception threshold on agents")
    parser.add_argument("--th_yaw_degree", type=float, default=TH_YAW_DEGREE, help="max absolute distance in degree")
    parser.add_argument("--th_extent_ratio", type=float, default=TH_EXTENT_RATIO, help="max change in area allowed")
    parser.add_argument("--th_distance_av", type=float, default=TH_DISTANCE_AV, help="max distance from AV in meters")
    args = parser.parse_args()

    for input_folder in args.input_folders:
        zarr_dataset = ChunkedDataset(path=input_folder)
        zarr_dataset.open()

        select_agents(
            zarr_dataset, args.th_agent_prob, args.th_yaw_degree, args.th_extent_ratio, args.th_distance_av,
        )



from enum import IntEnum
from typing import Callable, Optional, Tuple

import numpy as np
import torch

from l5kit.planning import utils


metric_signature = Callable[[np.ndarray, np.ndarray, np.ndarray, np.ndarray], np.ndarray]


def _assert_shapes(ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray) -> None:
    """
    Check the shapes of args required by metrics

    Args:
        ground_truth (np.ndarray): array of shape (timesteps)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(timesteps)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (timesteps) with the availability for each gt timesteps

    Returns:

    """
    assert len(pred.shape) == 3, f"expected 3D (MxTxC) array for pred, got {pred.shape}"
    num_modes, future_len, num_coords = pred.shape

    assert ground_truth.shape == (
        future_len,
        num_coords,
    ), f"expected 2D (Time x Coords) array for gt, got {ground_truth.shape}"
    assert confidences.shape == (num_modes,), f"expected 1D (Modes) array for confidences, got {confidences.shape}"
    assert np.allclose(np.sum(confidences), 1), "confidences should sum to 1"
    assert avails.shape == (future_len,), f"expected 1D (Time) array for avails, got {avails.shape}"
    # assert all data are valid
    assert np.isfinite(pred).all(), "invalid value found in pred"
    assert np.isfinite(ground_truth).all(), "invalid value found in gt"
    assert np.isfinite(confidences).all(), "invalid value found in confidences"
    assert np.isfinite(avails).all(), "invalid value found in avails"


def neg_multi_log_likelihood(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray
) -> np.ndarray:
    """
    Compute a negative log-likelihood for the multi-modal scenario.
    log-sum-exp trick is used here to avoid underflow and overflow, For more information about it see:
    https://en.wikipedia.org/wiki/LogSumExp#log-sum-exp_trick_for_log-domain_calculations
    https://timvieira.github.io/blog/post/2014/02/11/exp-normalize-trick/
    https://leimao.github.io/blog/LogSumExp/

    For more details about used loss function and reformulation, please see
    https://github.com/lyft/l5kit/blob/master/competition.md.

    Args:
        ground_truth (np.ndarray): array of shape (timesteps)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(timesteps)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (timesteps) with the availability for each gt timesteps

    Returns:
        np.ndarray: negative log-likelihood for this example, a single float number
    """
    _assert_shapes(ground_truth, pred, confidences, avails)

    ground_truth = np.expand_dims(ground_truth, 0)  # add modes
    avails = avails[np.newaxis, :, np.newaxis]  # add modes and cords

    error = np.sum(((ground_truth - pred) * avails) ** 2, axis=-1)  # reduce coords and use availability

    with np.errstate(divide="ignore"):  # when confidence is 0 log goes to -inf, but we're fine with it
        error = np.log(confidences) - 0.5 * np.sum(error, axis=-1)  # reduce timesteps

    # use max aggregator on modes for numerical stability
    max_value = error.max()  # error are negative at this point, so max() gives the minimum one
    error = -np.log(np.sum(np.exp(error - max_value), axis=-1)) - max_value  # reduce modes
    return error


def rmse(ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray) -> np.ndarray:
    """
    Return the root mean squared error, computed using the stable nll

    Args:
        ground_truth (np.ndarray): array of shape (timesteps)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(timesteps)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (timesteps) with the availability for each gt timesteps

    Returns:
        np.ndarray: negative log-likelihood for this example, a single float number

    """
    nll = neg_multi_log_likelihood(ground_truth, pred, confidences, avails)
    _, future_len, _ = pred.shape

    return np.sqrt(2 * nll / future_len)


def prob_true_mode(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray
) -> np.ndarray:
    """
    Return the probability of the true mode

    Args:
        ground_truth (np.ndarray): array of shape (timesteps)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(timesteps)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (timesteps) with the availability for each gt timesteps

    Returns:
        np.ndarray: a (modes) numpy array

    """
    _assert_shapes(ground_truth, pred, confidences, avails)

    ground_truth = np.expand_dims(ground_truth, 0)  # add modes
    avails = avails[np.newaxis, :, np.newaxis]  # add modes and cords

    error = np.sum(((ground_truth - pred) * avails) ** 2, axis=-1)  # reduce coords and use availability

    with np.errstate(divide="ignore"):  # when confidence is 0 log goes to -inf, but we're fine with it
        error = np.log(confidences) - 0.5 * np.sum(error, axis=-1)  # reduce timesteps

    # use max aggregator on modes for numerical stability
    max_value = error.max()  # error are negative at this point, so max() gives the minimum one

    error = np.exp(error - max_value) / np.sum(np.exp(error - max_value))
    return error


def time_displace(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray
) -> np.ndarray:
    """
    Return the displacement at timesteps T

    Args:
        ground_truth (np.ndarray): array of shape (timesteps)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(timesteps)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (timesteps) with the availability for each gt timesteps

    Returns:
        np.ndarray: a (timesteps) numpy array

    """
    true_mode_error = prob_true_mode(ground_truth, pred, confidences, avails)
    true_mode_error = true_mode_error[:, None]  # add timesteps axis

    ground_truth = np.expand_dims(ground_truth, 0)  # add modes
    avails = avails[np.newaxis, :, np.newaxis]  # add modes and cords

    error = np.sum(((ground_truth - pred) * avails) ** 2, axis=-1)  # reduce coords and use availability
    return np.sum(true_mode_error * np.sqrt(error), axis=0)  # reduce modes


def _average_displacement_error(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray, mode: str
) -> np.ndarray:
    """
    Returns the average displacement error (ADE), which is the average displacement over all timesteps.
    During calculation, confidences are ignored, and two modes are available:
        - oracle: only consider the best hypothesis
        - mean: average over all hypotheses

    Args:
        ground_truth (np.ndarray): array of shape (time)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(time)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (time) with the availability for each gt timestep
        mode (str): calculation mode - options are 'mean' (average over hypotheses) and 'oracle' (use best hypotheses)

    Returns:
        np.ndarray: average displacement error (ADE), a single float number
    """
    _assert_shapes(ground_truth, pred, confidences, avails)

    ground_truth = np.expand_dims(ground_truth, 0)  # add modes
    avails = avails[np.newaxis, :, np.newaxis]  # add modes and cords

    error = np.sum(((ground_truth - pred) * avails) ** 2, axis=-1)  # reduce coords and use availability
    error = error ** 0.5  # calculate root of error (= L2 norm)
    error = np.mean(error, axis=-1)  # average over timesteps

    if mode == "oracle":
        error = np.min(error)  # use best hypothesis
    elif mode == "mean":
        error = np.mean(error, axis=0)  # average over hypotheses
    else:
        raise ValueError(f"mode: {mode} not valid")

    return error


def average_displacement_error_oracle(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray,
) -> np.ndarray:
    """
    Calls _average_displacement_error() to get the oracle average displacement error.

    Args:
        ground_truth (np.ndarray): array of shape (time)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(time)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (time) with the availability for each gt timestep

    Returns:
        np.ndarray: oracle average displacement error (ADE), a single float number
    """

    return _average_displacement_error(ground_truth, pred, confidences, avails, "oracle")


def average_displacement_error_mean(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray,
) -> np.ndarray:
    """
    Calls _average_displacement_error() to get the mean average displacement error.

    Args:
        ground_truth (np.ndarray): array of shape (time)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(time)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (time) with the availability for each gt timestep

    Returns:
        np.ndarray: mean average displacement error (ADE), a single float number
    """

    return _average_displacement_error(ground_truth, pred, confidences, avails, "mean")


def _final_displacement_error(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray, mode: str
) -> np.ndarray:
    """
    Returns the final displacement error (FDE), which is the displacement calculated at the last timestep.
    During calculation, confidences are ignored, and two modes are available:
        - oracle: only consider the best hypothesis
        - mean: average over all hypotheses

    Args:
        ground_truth (np.ndarray): array of shape (time)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(time)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (time) with the availability for each gt timestep
        mode (str): calculation mode - options are 'mean' (average over hypotheses) and 'oracle' (use best hypotheses)

    Returns:
        np.ndarray: final displacement error (FDE), a single float number
    """
    _assert_shapes(ground_truth, pred, confidences, avails)

    ground_truth = np.expand_dims(ground_truth, 0)  # add modes
    avails = avails[np.newaxis, :, np.newaxis]  # add modes and cords

    error = np.sum(((ground_truth - pred) * avails) ** 2, axis=-1)  # reduce coords and use availability
    error = error ** 0.5  # calculate root of error (= L2 norm)
    error = error[:, -1]  # use last timestep

    if mode == "oracle":
        error = np.min(error)  # use best hypothesis
    elif mode == "mean":
        error = np.mean(error, axis=0)  # average over hypotheses
    else:
        raise ValueError(f"mode: {mode} not valid")

    return error


def final_displacement_error_oracle(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray,
) -> np.ndarray:
    """
    Calls _final_displacement_error() to get the oracle average displacement error.

    Args:
        ground_truth (np.ndarray): array of shape (time)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(time)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (time) with the availability for each gt timestep

    Returns:
        np.ndarray: oracle final displacement error (FDE), a single float number
    """

    return _final_displacement_error(ground_truth, pred, confidences, avails, "oracle")


def final_displacement_error_mean(
        ground_truth: np.ndarray, pred: np.ndarray, confidences: np.ndarray, avails: np.ndarray,
) -> np.ndarray:
    """
    Calls _final_displacement_error() to get the mean average displacement error.

    Args:
        ground_truth (np.ndarray): array of shape (time)x(2D coords)
        pred (np.ndarray): array of shape (modes)x(time)x(2D coords)
        confidences (np.ndarray): array of shape (modes) with a confidence for each mode in each sample
        avails (np.ndarray): array of shape (time) with the availability for each gt timestep

    Returns:
        np.ndarray: mean final displacement error (FDE), a single float number
    """

    return _final_displacement_error(ground_truth, pred, confidences, avails, "mean")


class CollisionType(IntEnum):
    """This enum defines the three types of collisions: front, rear and side."""
    FRONT = 0
    REAR = 1
    SIDE = 2


def detect_collision(pred_centroid: np.ndarray, pred_yaw: np.ndarray,
                     pred_extent: np.ndarray, target_agents: np.ndarray) -> Optional[Tuple[CollisionType, str]]:
    """
    Computes whether a collision occured between ego and any another agent.
    Also computes the type of collision: rear, front, or side.
    For this, we compute the intersection of ego's four sides with a target
    agent and measure the length of this intersection. A collision
    is classified into a class, if the corresponding length is maximal,
    i.e. a front collision exhibits the longest intersection with
    egos front edge.

    .. note:: please note that this funciton will stop upon finding the first
              colision, so it won't return all collisions but only the first
              one found.

    :param pred_centroid: predicted centroid
    :param pred_yaw: predicted yaw
    :param pred_extent: predicted extent
    :param target_agents: target agents
    :return: None if not collision was found, and a tuple with the
             collision type and the agent track_id
    """
    ego_bbox = utils._get_bounding_box(centroid=pred_centroid, yaw=pred_yaw, extent=pred_extent)
    within_range_mask = utils.within_range(pred_centroid, pred_extent,
                                           target_agents["centroid"], target_agents["extent"])
    for agent in target_agents[within_range_mask]:
        agent_bbox = utils._get_bounding_box(agent["centroid"], agent["yaw"], agent["extent"])

        if ego_bbox.intersects(agent_bbox):
            front_side, rear_side, left_side, right_side = utils._get_sides(ego_bbox)

            intersection_length_per_side = np.asarray(
                [
                    agent_bbox.intersection(front_side).length,
                    agent_bbox.intersection(rear_side).length,
                    agent_bbox.intersection(left_side).length,
                    agent_bbox.intersection(right_side).length,
                ]
            )
            argmax_side = np.argmax(intersection_length_per_side)

            # Remap here is needed because there are two sides that are
            # mapped to the same collision type CollisionType.SIDE
            max_collision_types = max(CollisionType).value
            remap_argmax = min(argmax_side, max_collision_types)
            collision_type = CollisionType(remap_argmax)
            return collision_type, agent["track_id"]
    return None


def distance_to_reference_trajectory(pred_centroid: torch.Tensor, ref_traj: torch.Tensor) -> torch.Tensor:
    """ Computes the distance from the predicted centroid to the closest waypoint in the reference trajectory.

    :param pred_centroid: predicted centroid tensor, size: [batch_size, 2]
    :type pred_centroid: torch.Tensor, float
    :param ref_traj: reference trajectory tensor, size: [batch_size, num_timestamps, 2]
    :type ref_traj: torch.Tensor, float
    :return: closest distance between the predicted centroid and the reference trajectory, size: [batch_size,]
    :rtype: torch.Tensor, float
    """
    # [batch_size, 2]
    assert pred_centroid.dim() == 2
    # [batch_size, num_timestamps, 2]
    assert ref_traj.dim() == 3

    # [batch_size,]
    euclidean_distance = torch.linalg.norm(pred_centroid.unsqueeze(1) - ref_traj, ord=2, dim=-1)
    return torch.amin(euclidean_distance, dim=1)
